"""Methods to view database items

"""
import os.path
import webbrowser

import keepmenu
from keepmenu.menu import dmenu_select
from keepmenu.totp import gen_otp, get_otp_url


def view_all_entries(options, kp_entries, dbname):
    """Generate numbered list of all Keepass entries and open with dmenu.

    Returns: dmenu selection

    """
    num_align = len(str(len(kp_entries)))
    kp_entry_pattern = str("{:>{na}} - {} - {} - {}")  # Path,username,url
    # Have to number each entry to capture duplicates correctly
    kps = str("\n").join([kp_entry_pattern.format(j,
                                                  os.path.join("/".join(i.path[:-1]),
                                                               i.deref('title') or ""),
                                                  i.deref('username'),
                                                  i.deref('url'),
                                                  na=num_align)
                         for j, i in enumerate(kp_entries)])
    if options:
        options_s = ("\n".join(options) + "\n")
        entries_s = options_s + kps
    else:
        entries_s = kps

    prompt = f"Entries: {dbname}"
    if keepmenu.CONF.has_option('dmenu', 'title_path'):
        try:
            max_length = keepmenu.CONF.getboolean('dmenu', 'title_path')
        except ValueError:
            max_length = keepmenu.CONF.getint('dmenu', 'title_path')
        prompt = generate_prompt(max_length, dbname)

    return dmenu_select(min(keepmenu.MAX_LEN, len(options) + len(kp_entries)),
                        inp=entries_s,
                        prompt=prompt)


def view_entry(kp_entry):
    """Show title, username, password, url and notes for an entry.

    Returns: dmenu selection

    """
    fields = [os.path.join("/".join(kp_entry.path[:-1]), kp_entry.deref('title') or "")
              or "Title: None",
              kp_entry.deref('username') or "Username: None",
              '**********' if kp_entry.deref('password') else "Password: None",
              "TOTP: ******" if get_otp_url(kp_entry) else "TOTP: None",
              kp_entry.deref('url') or "URL: None",
              "Notes: <Enter to view>" if kp_entry.deref('notes') else "Notes: None",
              str(f"Expire time: {kp_entry.expiry_time}")
              if kp_entry.expires is True else "Expiry date: None"]

    attrs = kp_entry.custom_properties
    for attr in attrs:
        if attr != "otp":
            val = attrs.get(attr) or ""
            value = val or "None" if len(val.split('\n')) <= 1 else "<Enter to view>"
            fields.append(f'{attr}: {value}')

    sel = dmenu_select(len(fields), inp="\n".join(fields))
    if sel == "Notes: <Enter to view>":
        sel = view_notes(kp_entry.deref('notes'))
    elif sel == "Notes: None":
        sel = ""
    elif sel == '**********':
        sel = kp_entry.deref('password')
    elif sel == "TOTP: ******":
        sel = gen_otp(get_otp_url(kp_entry))
    elif sel == fields[4]:
        if sel != "URL: None":
            webbrowser.open(sel)
        sel = ""
    else:
        for attr in attrs:
            if sel == f'{attr}: {attrs.get(attr) or ""}':
                sel = attrs.get(attr)
                break
            if sel == f'{attr}: <Enter to view>':
                sel = view_notes(attrs.get(attr) or "")

    return sel if not sel.endswith(": None") else ""


def view_notes(notes):
    """View the 'Notes' field line-by-line within dmenu.

    Returns: text of the selected line for typing

    """
    notes_l = notes.split('\n')
    sel = dmenu_select(min(keepmenu.MAX_LEN, len(notes_l)), inp=notes)
    return sel


def generate_prompt(max_length, dbname):
    """Generate a prompt in the format "Entries: {}", with "{}" replaced by
    the full path to the database truncated to a certain length

    max_length: an int giving the maximum length for the path, or a bool
    specifying whether to show the entire path (True) or to hide it (False)

    dbname: the full path to the database

    """
    if max_length is False or max_length == 0:
        return "Entries"
    if max_length is True or max_length is None:
        return f"Entries: {dbname}"
    # Truncate the path so that it is no more than max_length
    # or the length of the filename, whichever is larger
    filename = os.path.basename(dbname)
    if len(filename) >= max_length - 3:
        return f"Entries: {filename}"
    path = dbname.replace(os.path.expanduser("~"), "~")
    if len(path) <= max_length:
        return f"Entries: {path}"
    path = path[:(max_length - len(filename) - 3)]
    return f"Entries: {path}...{filename}"



"""Set global variables. Read the config file. Create default config file if one
doesn't exist.

"""
import configparser
import locale
import os
import shlex
from subprocess import run, DEVNULL
import sys
from os.path import exists, expanduser

from keepmenu.menu import dmenu_err

# Setup logging for debugging. Usage: logger.info(...)
# import logging
# logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)
# file_handler = logging.FileHandler('keepmenu.log', mode='w')
# formatter = logging.Formatter('%(message)s')
# file_handler.setFormatter(formatter)
# logger.addHandler(file_handler)

AUTH_FILE = expanduser("~/.cache/.keepmenu-auth")
CONF_FILE = expanduser("~/.config/keepmenu/config.ini")
SECRET_VALID_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"

ENV = os.environ.copy()
ENV['LC_ALL'] = 'C'
ENC = locale.getpreferredencoding()
CACHE_PERIOD_DEFAULT_MIN = 360
CACHE_PERIOD_MIN = CACHE_PERIOD_DEFAULT_MIN
SEQUENCE = "{USERNAME}{TAB}{PASSWORD}{ENTER}"
MAX_LEN = 24
CONF = configparser.ConfigParser()


def reload_config():  # pylint: disable=too-many-statements,too-many-branches
    """Reload config file. Primarly for use with tests.

    """
    # pragma pylint: disable=global-statement,global-variable-not-assigned
    global CACHE_PERIOD_MIN, \
        CACHE_PERIOD_DEFAULT_MIN, \
        CONF, \
        MAX_LEN, \
        ENV, \
        ENC, \
        SEQUENCE
    # pragma pylint: enable=global-variable-undefined,global-variable-not-assigned
    CONF = configparser.ConfigParser()
    if not exists(CONF_FILE):
        try:
            os.mkdir(os.path.dirname(CONF_FILE))
        except OSError:
            pass
        with open(CONF_FILE, 'w', encoding=ENC) as conf_file:
            CONF.add_section('dmenu')
            CONF.set('dmenu', 'dmenu_command', 'dmenu')
            CONF.add_section('dmenu_passphrase')
            CONF.set('dmenu_passphrase', 'obscure', 'True')
            CONF.set('dmenu_passphrase', 'obscure_color', '#222222')
            CONF.add_section('database')
            CONF.set('database', 'database_1', '')
            CONF.set('database', 'keyfile_1', '')
            CONF.set('database', 'pw_cache_period_min', str(CACHE_PERIOD_DEFAULT_MIN))
            CONF.set('database', 'autotype_default', SEQUENCE)
            CONF.write(conf_file)
    try:
        CONF.read(CONF_FILE)
    except configparser.ParsingError as err:
        dmenu_err(f"Config file error: {err}")
        sys.exit()
    if CONF.has_option('dmenu', 'dmenu_command'):
        command = shlex.split(CONF.get('dmenu', 'dmenu_command'))
    else:
        CONF.set('dmenu', 'dmenu_command', 'dmenu')
        command = 'dmenu'
    if "-l" in command:
        MAX_LEN = int(command[command.index("-l") + 1])
    if CONF.has_option("database", "pw_cache_period_min"):
        CACHE_PERIOD_MIN = int(CONF.get("database", "pw_cache_period_min"))
    else:
        CACHE_PERIOD_MIN = CACHE_PERIOD_DEFAULT_MIN
    if CONF.has_option('database', 'autotype_default'):
        SEQUENCE = CONF.get("database", "autotype_default")
    if CONF.has_option("database", "type_library"):
        if CONF.get("database", "type_library") == "xdotool":
            try:
                run(['xdotool', 'version'], check=False, stdout=DEVNULL)
            except OSError:
                dmenu_err("Xdotool not installed.\n"
                          "Please install or remove that option from config.ini")
                sys.exit()
        elif CONF.get("database", "type_library") == "ydotool":
            try:
                run(['ydotool'], check=False, stdout=DEVNULL)
            except OSError:
                dmenu_err("Ydotool not installed.\n"
                          "Please install or remove that option from config.ini")
                sys.exit()
        elif CONF.get("database", "type_library") == "wtype":
            try:
                run(['wtype'], check=False, stdout=DEVNULL, stderr=DEVNULL)
            except OSError:
                dmenu_err("wtype not installed.\n"
                          "Please install or remove that option from config.ini")
                sys.exit()


reload_config()

# vim: set et ts=4 sw=4 :



"""Methods for editing entries and groups

"""
import os
import random
from secrets import choice
import shlex
import string
from subprocess import call
import tempfile
from urllib import parse

import keepmenu
from keepmenu.menu import dmenu_select, dmenu_err
from keepmenu.totp import gen_otp, get_otp_url
from keepmenu.type import type_text


def add_entry(kpo):
    """Add Keepass entry

    Args: kpo - Keepass object
    Returns: False if not added
             Keepass Entry object on success

    """
    group = select_group(kpo)
    if group is False:
        return False
    entry = kpo.add_entry(destination_group=group, title="", username="", password="")
    edit = True
    while edit is True:
        edit = edit_entry(kpo, entry)
    return entry


def delete_entry(kpo, kp_entry):
    """Delete an entry

    Args: kpo - Keepass object
          kp_entry - keepass entry
    Returns: True if no delete
             'del' if delete

    """
    inp = "NO\nYes - confirm delete\n"
    delete = dmenu_select(2, "Confirm delete", inp=inp)
    if delete != "Yes - confirm delete":
        return True
    kpo.delete_entry(kp_entry)
    kpo.save()
    return 'del'


def edit_entry(kpo, kp_entry):
    # pylint: disable=too-many-return-statements,too-many-branches,too-many-statements
    """Edit title, username, password, url and autotype sequence for an entry.

    Args: kpo - Keepass object
          kp_entry - selected Entry object

    Returns: True to continue editing
             False if done
             "del" if entry was deleted

    """
    fields = [str(f"Title: {kp_entry.title}"),
              str(f"Path: {'/'.join(kp_entry.path[:-1])}"),
              str(f"Username: {kp_entry.username}"),
              str("Password: **********") if kp_entry.password else "Password: None",
              str("TOTP: ******") if get_otp_url(kp_entry) else "TOTP: None",
              str(f"Url: {kp_entry.url}"),
              "Notes: <Enter to Edit>" if kp_entry.notes else "Notes: None",
              str(f"Expiry time: {kp_entry.expiry_time}")
              if kp_entry.expires is True else "Expiry date: None"]

    attrs = kp_entry.custom_properties
    for attr in attrs:
        if attr != "otp":
            val = attrs.get(attr) or ""
            value = val or "None" if len(val.split('\n')) <= 1 else "<Enter to Edit>"
            fields.append(f'{attr}: {value}')

    fields.append("Add Attribute: ")
    fields.append("Delete Entry: ")

    if hasattr(kp_entry, 'autotype_sequence') and hasattr(kp_entry, 'autotype_enabled'):
        fields[5:5] = [str(f"Autotype Sequence: {kp_entry.autotype_sequence}"),
                       str(f"Autotype Enabled: {kp_entry.autotype_enabled}")]
    inp = "\n".join(fields)
    sel = dmenu_select(len(fields), inp=inp)

    # Needs to be above so that underscores are not added to the key
    for attr in attrs:
        if sel in (f'{attr}: {attrs.get(attr) or "None"}', f'{attr}: <Enter to Edit>'):
            edit_additional_attributes(kp_entry, attr, attrs.get(attr) or "")
            return True

    try:
        field, sel = sel.split(": ", 1)
    except (ValueError, TypeError):
        return False
    field = field.lower().replace(" ", "_")
    if field == 'password':
        sel = kp_entry.password
    edit = f"{sel}\n" if sel is not None else "\n"
    if field == 'delete_entry':
        return delete_entry(kpo, kp_entry)
    if field == 'path':
        group = select_group(kpo)
        if not group:
            return True
        kpo.move_entry(kp_entry, group)
        return True
    pw_choice = ""
    if field == 'password':
        inputs = [
            "Generate password",
            "Manually enter password",
        ]
        if kp_entry.password:
            inputs.append("Type existing password")
        pw_choice = dmenu_select(len(inputs), "Password Options", inp="\n".join(inputs))
        if pw_choice == "Manually enter password":
            pass
        elif pw_choice == "Type existing password":
            type_text(kp_entry.password)
            return False
        elif not pw_choice:
            return True
        else:
            pw_choice = ''
            length = dmenu_select(1, "Password Length?", inp="20\n")
            if not length:
                return True
            try:
                length = int(length)
            except ValueError:
                length = 20
            chars = get_password_chars()
            if chars is False:
                return True
            sel = gen_passwd(chars, length)
            if sel is False:
                dmenu_err("Number of char groups desired is more than requested pw length")
                return True
    if field == 'totp':
        edit_totp(kp_entry)
        return True

    if field == "add_attribute":
        add_additional_attribute(kp_entry)
        return True

    if field == 'autotype_enabled':
        inp = "True\nFalse\n"
        at_enab = dmenu_select(2, "Autotype Enabled? True/False", inp=inp)
        if not at_enab:
            return True
        sel = not at_enab == 'False'

    if (field not in ('password', 'notes', 'path', 'autotype_enabled')) or pw_choice:
        sel = dmenu_select(1, f"{field.capitalize()}", inp=edit)
        if not sel:
            return True
        if pw_choice:
            sel_check = dmenu_select(1, f"{field.capitalize()}", inp=edit)
            if not sel_check or sel_check != sel:
                dmenu_err("Passwords do not match. No changes made.")
                return True
    elif field == 'notes':
        sel = edit_notes(kp_entry.notes)

    setattr(kp_entry, field, sel)
    return True


def edit_totp(kp_entry):  # pylint: disable=too-many-statements,too-many-branches
    """Edit TOTP generation information

    Args: kp_entry - selected Entry object

    """
    otp_url = get_otp_url(kp_entry)

    if otp_url is not None:
        inputs = [
            "Enter secret key",
            "Type TOTP",
        ]
        otp_choice = dmenu_select(len(inputs), "TOTP", inp="\n".join(inputs))
    else:
        otp_choice = "Enter secret key"

    if otp_choice == "Type TOTP":
        type_text(gen_otp(otp_url))
    elif otp_choice == "Enter secret key":
        inputs = []
        if otp_url:
            parsed_otp_url = parse.urlparse(otp_url)
            query_string = parse.parse_qs(parsed_otp_url.query)
            inputs = [query_string["secret"][0]]
        secret_key = dmenu_select(1, "Secret Key?", inp="\n".join(inputs))

        if not secret_key:
            return

        for char in secret_key:
            if char.upper() not in keepmenu.SECRET_VALID_CHARS:
                dmenu_err("Invaild character in secret key, "
                          f"valid characters are {keepmenu.SECRET_VALID_CHARS}")
                return

        inputs = [
            "Defaut RFC 6238 token settings",
            "Steam token settings",
            "Use cusom settings"
        ]

        otp_settings_choice = dmenu_select(len(inputs), "Settings", inp="\n".join(inputs))

        if otp_settings_choice == "Defaut RFC 6238 token settings":
            algorithm_choice = "sha1"
            time_step_choice = 30
            code_size_choice = 6
        elif otp_settings_choice == "Steam token settings":
            algorithm_choice = "sha1"
            time_step_choice = 30
            code_size_choice = 5
        elif otp_settings_choice == "Use custom settings":
            inputs = ["SHA-1", "SHA-256", "SHA-512"]
            algorithm_choice = dmenu_select(len(inputs), "Algorithm", inp="\n".join(inputs))
            if not algorithm_choice:
                return
            algorithm_choice = algorithm_choice.replace("-", "").lower()

            time_step_choice = dmenu_select(1, "Time Step (sec)", inp="30\n")
            if not time_step_choice:
                return
            try:
                time_step_choice = int(time_step_choice)
            except ValueError:
                time_step_choice = 30

            code_size_choice = dmenu_select(1, "Code Size", inp="6\n")
            if not code_size_choice:
                return
            try:
                code_size_choice = int(time_step_choice)
            except ValueError:
                code_size_choice = 6

        otp_url = (f"otpauth://totp/Main:none?secret={secret_key}&period={time_step_choice}"
                   f"&digits={code_size_choice}&issuer=Main")
        if algorithm_choice != "sha1":
            otp_url += "&algorithm=" + algorithm_choice
        if otp_settings_choice == "Steam token settings":
            otp_url += "&encoder=steam"

        if hasattr(kp_entry, "otp"):
            setattr(kp_entry, "otp", otp_url)
        else:
            kp_entry.set_custom_property("otp", otp_url)


def add_additional_attribute(kp_entry):
    """Add additional attribute

        Args: kp_entry - Entry object

    """
    key = dmenu_select(0, "Attribute Name: ", inp="")
    if not key:
        return
    edit_additional_attributes(kp_entry, key, "")


def edit_additional_attributes(kp_entry, key, value):
    """Edit/Delete additional attributes

    Args: kp_entry - Entry object
          key - attr name (string)
          value - attr value (string)

    """
    options = ["Multi-line Edit", "Delete"]
    if len(value.split('\n')) == 1:
        options.insert(0, value)
    sel = dmenu_select(len(options), f"{key}:", inp="\n".join(options))
    if sel == "Delete":
        kp_entry.delete_custom_property(key)
        return
    if sel == "Multi-line Edit":
        sel = edit_notes(value)
    if sel:
        kp_entry.set_custom_property(key, sel)


def edit_notes(note):
    """Use $EDITOR (or 'vim' if not set) to edit the notes entry

    In configuration file:
        Set 'gui_editor' for things like emacs, gvim, leafpad
        Set 'editor' for vim, emacs -nw, nano unless $EDITOR is defined
        Set 'terminal' if using a non-gui editor

    Args: note - string
    Returns: note - string

    """
    if keepmenu.CONF.has_option("database", "gui_editor"):
        editor = keepmenu.CONF.get("database", "gui_editor")
        editor = shlex.split(editor)
    else:
        if keepmenu.CONF.has_option("database", "editor"):
            editor = keepmenu.CONF.get("database", "editor")
        else:
            editor = os.environ.get('EDITOR', 'vim')
        if keepmenu.CONF.has_option("database", "terminal"):
            terminal = keepmenu.CONF.get("database", "terminal")
        else:
            terminal = "xterm"
        terminal = shlex.split(terminal)
        editor = shlex.split(editor)
        editor = terminal + ["-e"] + editor
    note = b'' if note is None else note.encode(keepmenu.ENC)
    with tempfile.NamedTemporaryFile(suffix=".tmp") as fname:
        fname.write(note)
        fname.flush()
        editor.append(fname.name)
        try:
            call(editor)
        except FileNotFoundError:
            dmenu_err("Terminal not found. Please update config.ini.")
            note = '' if not note else note.decode(keepmenu.ENC)
            return note
        fname.seek(0)
        note = fname.read()
    note = '' if not note else note.decode(keepmenu.ENC)
    return note.strip()


def gen_passwd(chars, length=20):
    """Generate password (min = # of distinct character sets picked)

    Args: chars - Dict {preset_name_1: {char_set_1: string, char_set_2: string},
                        preset_name_2: ....}
          length - int (default 20)

    Returns: password - string OR False

    """
    sets = set()
    if chars:
        sets = set(j for i in chars.values() for j in i.values())
    if length < len(sets) or not chars:
        return False
    alphabet = "".join(set("".join(j for j in i.values()) for i in chars.values()))
    # Ensure minimum of one char from each character set
    password = "".join(choice(k) for k in sets)
    password += "".join(choice(alphabet) for i in range(length - len(sets)))
    tpw = list(password)
    random.shuffle(tpw)
    return "".join(tpw)


def get_password_chars():
    """Get characters to use for password generation from defaults, config file
    and user input.

    Returns: Dict {preset_name_1: {char_set_1: string, char_set_2: string},
                   preset_name_2: ....}
    """
    chars = {"upper": string.ascii_uppercase,
             "lower": string.ascii_lowercase,
             "digits": string.digits,
             "punctuation": string.punctuation}
    presets = {}
    presets["Letters+Digits+Punctuation"] = chars
    presets["Letters+Digits"] = {k: chars[k] for k in ("upper", "lower", "digits")}
    presets["Letters"] = {k: chars[k] for k in ("upper", "lower")}
    presets["Digits"] = {k: chars[k] for k in ("digits",)}
    if keepmenu.CONF.has_section('password_chars'):
        pw_chars = dict(keepmenu.CONF.items('password_chars'))
        chars.update(pw_chars)
        for key, val in pw_chars.items():
            presets[key.title()] = {k: chars[k] for k in (key,)}
    if keepmenu.CONF.has_section('password_char_presets'):
        if keepmenu.CONF.options('password_char_presets'):
            presets = {}
        for name, val in keepmenu.CONF.items('password_char_presets'):
            try:
                presets[name.title()] = {k: chars[k] for k in shlex.split(val)}
            except KeyError:
                print(f"Error: Unknown value in preset {name}. Ignoring.")
                continue
    inp = "\n".join(presets)
    char_sel = dmenu_select(len(presets),
                            "Pick character set(s) to use", inp=inp)
    # This dictionary return also handles Rofi multiple select
    return {k: presets[k] for k in char_sel.split('\n')} if char_sel else False


def select_group(kpo, prompt="Groups"):
    """Select which group for an entry

    Args: kpo - Keepass object
          options - list of menu options for groups

    Returns: False for no entry
             group - string

    """
    groups = kpo.groups
    num_align = len(str(len(groups)))
    pattern = str("{:>{na}} - {}")
    inp = str("\n").join([pattern.format(j, "/".join(i.path), na=num_align)
                         for j, i in enumerate(groups)])
    sel = dmenu_select(min(keepmenu.MAX_LEN, len(groups)), prompt, inp=inp)
    if not sel:
        return False
    try:
        return groups[int(sel.split('-', 1)[0])]
    except (ValueError, TypeError):
        return False


def manage_groups(kpo):
    """Rename, create, move or delete groups

    Args: kpo - Keepass object
    Returns: Group object or False

    """
    edit = True
    options = ['Create',
               'Move',
               'Rename',
               'Delete']
    group = False
    while edit is True:
        inp = "\n".join(i for i in options) + "\n\n" + \
            "\n".join("/".join(i.path) for i in kpo.groups)
        sel = dmenu_select(len(options) + len(kpo.groups) + 1, "Groups", inp=inp)
        if not sel:
            edit = False
        elif sel == 'Create':
            group = create_group(kpo)
        elif sel == 'Move':
            group = move_group(kpo)
        elif sel == 'Rename':
            group = rename_group(kpo)
        elif sel == 'Delete':
            group = delete_group(kpo)
        else:
            edit = False
    return group


def create_group(kpo):
    """Create new group

    Args: kpo - Keepass object
    Returns: Group object or False

    """
    parentgroup = select_group(kpo, prompt="Select parent group")
    if not parentgroup:
        return False
    name = dmenu_select(1, "Group name")
    if not name:
        return False
    group = kpo.add_group(parentgroup, name)
    kpo.save()
    return group


def delete_group(kpo):
    """Delete a group

    Args: kpo - Keepass object
    Returns: Group object or False

    """
    group = select_group(kpo, prompt="Delete Group:")
    if not group:
        return False
    inp = "NO\nYes - confirm delete\n"
    delete = dmenu_select(2, "Confirm delete", inp=inp)
    if delete != "Yes - confirm delete":
        return True
    kpo.delete_group(group)
    kpo.save()
    return group


def move_group(kpo):
    """Move group

    Args: kpo - Keepass object
    Returns: Group object or False

    """
    group = select_group(kpo, prompt="Select group to move")
    if not group:
        return False
    destgroup = select_group(kpo, prompt="Select destination group")
    if not destgroup:
        return False
    group = kpo.move_group(group, destgroup)
    kpo.save()
    return group


def rename_group(kpo):
    """Rename group

    Args: kpo - Keepass object
    Returns: Group object or False

    """
    group = select_group(kpo, prompt="Select group to rename")
    if not group:
        return False
    name = dmenu_select(1, "New group name", inp=group.name)
    if not name:
        return False
    group.name = name
    kpo.save()
    return group



# flake8: noqa
from pynput import keyboard

AUTOTYPE_TOKENS = {
    "{TAB}"       : keyboard.Key.tab,
    "{ENTER}"     : keyboard.Key.enter,
    "~"           : keyboard.Key.enter,
    "{UP}"        : keyboard.Key.up,
    "{DOWN}"      : keyboard.Key.down,
    "{LEFT}"      : keyboard.Key.left,
    "{RIGHT}"     : keyboard.Key.right,
    "{INSERT}"    : keyboard.Key.insert,
    "{INS}"       : keyboard.Key.insert,
    "{DELETE}"    : keyboard.Key.delete,
    "{DEL}"       : keyboard.Key.delete,
    "{HOME}"      : keyboard.Key.home,
    "{END}"       : keyboard.Key.end,
    "{PGUP}"      : keyboard.Key.page_up,
    "{PGDN}"      : keyboard.Key.page_down,
    "{SPACE}"     : keyboard.Key.space,
    "{BACKSPACE}" : keyboard.Key.backspace,
    "{BS}"        : keyboard.Key.backspace,
    "{BKSP}"      : keyboard.Key.backspace,
    "{BREAK}"     : keyboard.Key.pause,
    "{CAPSLOCK}"  : keyboard.Key.caps_lock,
    "{ESC}"       : keyboard.Key.esc,
    "{WIN}"       : keyboard.Key.cmd,
    "{LWIN}"      : keyboard.Key.cmd_l,
    "{RWIN}"      : keyboard.Key.cmd_r,
    # "{APPS}"    : keyboard.Key.
    # "{HELP}"    : keyboard.Key.
    "{NUMLOCK}"   : keyboard.Key.num_lock,
    "{PRTSC}"     : keyboard.Key.print_screen,
    "{SCROLLLOCK}": keyboard.Key.scroll_lock,
    "{F1}"        : keyboard.Key.f1,
    "{F2}"        : keyboard.Key.f2,
    "{F3}"        : keyboard.Key.f3,
    "{F4}"        : keyboard.Key.f4,
    "{F5}"        : keyboard.Key.f5,
    "{F6}"        : keyboard.Key.f6,
    "{F7}"        : keyboard.Key.f7,
    "{F8}"        : keyboard.Key.f8,
    "{F9}"        : keyboard.Key.f9,
    "{F10}"       : keyboard.Key.f10,
    "{F11}"       : keyboard.Key.f11,
    "{F12}"       : keyboard.Key.f12,
    "{F13}"       : keyboard.Key.f13,
    "{F14}"       : keyboard.Key.f14,
    "{F15}"       : keyboard.Key.f15,
    "{F16}"       : keyboard.Key.f16,
    # "{ADD}"       : keyboard.Key.
    # "{SUBTRACT}"  : keyboard.Key.
    # "{MULTIPLY}"  : keyboard.Key.
    # "{DIVIDE}"    : keyboard.Key.
    # "{NUMPAD0}"   : keyboard.Key.
    # "{NUMPAD1}"   : keyboard.Key.
    # "{NUMPAD2}"   : keyboard.Key.
    # "{NUMPAD3}"   : keyboard.Key.
    # "{NUMPAD4}"   : keyboard.Key.
    # "{NUMPAD5}"   : keyboard.Key.
    # "{NUMPAD6}"   : keyboard.Key.
    # "{NUMPAD7}"   : keyboard.Key.
    # "{NUMPAD8}"   : keyboard.Key.
    # "{NUMPAD9}"   : keyboard.Key.
    "+"           : keyboard.Key.shift,
    "^"           : keyboard.Key.ctrl,
    "%"           : keyboard.Key.alt,
    "@"           : keyboard.Key.cmd,
}



"""Methods for typing entries with pynput, xdotool, ydotool, wtype

"""
# flake8: noqa
# pylint: disable=import-outside-toplevel
import re
from subprocess import call
import time

import keepmenu
from keepmenu.menu import dmenu_err
from keepmenu.totp import gen_otp, get_otp_url


def tokenize_autotype(autotype):
    """Process the autotype sequence

    Args: autotype - string
    Returns: tokens - generator ((token, if_special_char T/F), ...)

    """
    while autotype:
        opening_idx = -1
        for char in "{+^%~@":
            idx = autotype.find(char)
            if idx != -1 and (opening_idx == -1 or idx < opening_idx):
                opening_idx = idx

        if opening_idx == -1:
            # found the end of the string without further opening braces or
            # other characters
            yield autotype, False
            return

        if opening_idx > 0:
            yield autotype[:opening_idx], False

        if autotype[opening_idx] in "+^%~@":
            yield autotype[opening_idx], True
            autotype = autotype[opening_idx + 1:]
            continue

        closing_idx = autotype.find('}')
        if closing_idx == -1:
            dmenu_err("Unable to find matching right brace (}) while" +
                      f"tokenizing auto-type string: {autotype}\n")
            return
        if closing_idx == opening_idx + 1 and closing_idx + 1 < len(autotype) \
                and autotype[closing_idx + 1] == '}':
            yield "{}}", True
            autotype = autotype[closing_idx + 2:]
            continue
        yield autotype[opening_idx:closing_idx + 1], True
        autotype = autotype[closing_idx + 1:]


def token_command(token):
    """When token denotes a special command, this function provides a callable
    implementing its behaviour.

    """
    cmd = None

    def _check_delay():
        match = re.match(r'{DELAY (\d+)}', token)
        if match:
            delay = match.group(1)
            nonlocal cmd
            cmd = lambda _, t=delay: time.sleep(int(t) / 1000)
            return True
        return False

    def _check_additional_attribute():
        match = re.match(r'{S:(.*)}', token)
        if match:
            attr = match.group(1)
            nonlocal cmd
            cmd = lambda e, a=attr: e.get_custom_property(a)
            return True
        return False

    if _check_delay():  # {DELAY x}
        return cmd

    if _check_additional_attribute():  # {S:<attr>}
        return cmd

    return None


def type_entry(entry, db_autotype=None):
    """Pick which library to use to type strings

    Defaults to pynput

    Args: entry - The entry to type
          db_autotype - the database specific autotype that overrides 'autotype_default'

    """
    sequence = keepmenu.SEQUENCE

    if hasattr(entry, 'autotype_enabled') and entry.autotype_enabled is False:
        dmenu_err("Autotype disabled for this entry")
        return
    if db_autotype is not None and db_autotype != '':
        sequence = db_autotype
    if hasattr(entry, 'autotype_sequence') and \
            entry.autotype_sequence is not None and \
            entry.autotype_sequence != 'None':
        sequence = entry.autotype_sequence
    tokens = tokenize_autotype(sequence)

    library = 'pynput'
    if keepmenu.CONF.has_option('database', 'type_library'):
        library = keepmenu.CONF.get('database', 'type_library')
    if library == 'xdotool':
        type_entry_xdotool(entry, tokens)
    elif library == 'ydotool':
        type_entry_ydotool(entry, tokens)
    elif library == 'wtype':
        type_entry_wtype(entry, tokens)
    else:
        type_entry_pynput(entry, tokens)


PLACEHOLDER_AUTOTYPE_TOKENS = {
    "{TITLE}"   : lambda e: e.deref('title'),
    "{USERNAME}": lambda e: e.deref('username'),
    "{URL}"     : lambda e: e.deref('url'),
    "{PASSWORD}": lambda e: e.deref('password'),
    "{NOTES}"   : lambda e: e.deref('notes'),
    "{TOTP}"    : lambda e: gen_otp(get_otp_url(e)),
}

STRING_AUTOTYPE_TOKENS = {
    "{PLUS}"      : '+',
    "{PERCENT}"   : '%',
    "{CARET}"     : '^',
    "{TILDE}"     : '~',
    "{LEFTPAREN}" : '(',
    "{RIGHTPAREN}": ')',
    "{LEFTBRACE}" : '{',
    "{RIGHTBRACE}": '}',
    "{AT}"        : '@',
    "{+}"         : '+',
    "{%}"         : '%',
    "{^}"         : '^',
    "{~}"         : '~',
    "{(}"         : '(',
    "{)}"         : ')',
    "{[}"         : '[',
    "{]}"         : ']',
    "{{}"         : '{',
    "{}}"         : '}',
}

def type_entry_pynput(entry, tokens):  # pylint: disable=too-many-branches
    """Use pynput to auto-type the selected entry

    """
    try:
        from pynput import keyboard
        from .tokens_pynput import AUTOTYPE_TOKENS
    except ModuleNotFoundError:
        return
    kbd = keyboard.Controller()
    enter_idx = True
    for token, special in tokens:
        if special:
            cmd = token_command(token)
            if callable(cmd):
                to_type = cmd(entry)  # pylint: disable=not-callable
                if to_type is not None:
                    try:
                        kbd.type(to_type)
                    except kbd.InvalidCharacterException:
                        dmenu_err("Unable to type string...bad character.\n"
                                  "Try setting `type_library = xdotool` in config.ini")
            elif token in PLACEHOLDER_AUTOTYPE_TOKENS:
                to_type = PLACEHOLDER_AUTOTYPE_TOKENS[token](entry)
                if to_type:
                    try:
                        kbd.type(to_type)
                    except kbd.InvalidCharacterException:
                        dmenu_err("Unable to type string...bad character.\n"
                                  "Try setting `type_library = xdotool` in config.ini")
                        return
            elif token in STRING_AUTOTYPE_TOKENS:
                to_type = STRING_AUTOTYPE_TOKENS[token]
                try:
                    kbd.type(to_type)
                except kbd.InvalidCharacterException:
                    dmenu_err("Unable to type string...bad character.\n"
                              "Try setting `type_library = xdotool` in config.ini")
                    return
            elif token in AUTOTYPE_TOKENS:
                to_tap = AUTOTYPE_TOKENS[token]
                kbd.tap(to_tap)
                # Add extra {ENTER} key tap for first instance of {ENTER}. It
                # doesn't get recognized for some reason.
                if enter_idx is True and token in ("{ENTER}", "~"):
                    kbd.tap(to_tap)
                    enter_idx = False
            else:
                dmenu_err(f"Unsupported auto-type token (pynput): \"{token}\"")
                return
        else:
            try:
                kbd.type(token)
            except kbd.InvalidCharacterException:
                dmenu_err("Unable to type string...bad character.\n"
                          "Try setting `type_library = xdotool` in config.ini")
                return


def type_entry_xdotool(entry, tokens):
    """Auto-type entry entry using xdotool

    """
    enter_idx = True
    from .tokens_xdotool import AUTOTYPE_TOKENS
    for token, special in tokens:
        if special:
            cmd = token_command(token)
            if callable(cmd):
                to_type = cmd(entry)  # pylint: disable=not-callable
                if to_type is not None:
                    call(['xdotool', 'type', '--', to_type])
            elif token in PLACEHOLDER_AUTOTYPE_TOKENS:
                to_type = PLACEHOLDER_AUTOTYPE_TOKENS[token](entry)
                if to_type:
                    call(['xdotool', 'type', '--', to_type])
            elif token in STRING_AUTOTYPE_TOKENS:
                to_type = STRING_AUTOTYPE_TOKENS[token]
                call(['xdotool', 'type', '--', to_type])
            elif token in AUTOTYPE_TOKENS:
                cmd = ['xdotool'] + AUTOTYPE_TOKENS[token]
                call(cmd)
                # Add extra {ENTER} key tap for first instance of {ENTER}. It
                # doesn't get recognized for some reason.
                if enter_idx is True and token in ("{ENTER}", "~"):
                    cmd = ['xdotool'] + AUTOTYPE_TOKENS[token]
                    call(cmd)
                    enter_idx = False
            else:
                dmenu_err(f"Unsupported auto-type token (xdotool): \"{token}\"")
                return
        else:
            call(['xdotool', 'type', '--', token])


def type_entry_ydotool(entry, tokens):
    """Auto-type entry entry using ydotool

    """
    from .tokens_ydotool import AUTOTYPE_TOKENS
    for token, special in tokens:
        if special:
            cmd = token_command(token)
            if callable(cmd):
                to_type = cmd(entry)  # pylint: disable=not-callable
                if to_type is not None:
                    call(['ydotool', 'type', '--', to_type])
            elif token in PLACEHOLDER_AUTOTYPE_TOKENS:
                to_type = PLACEHOLDER_AUTOTYPE_TOKENS[token](entry)
                if to_type:
                    call(['ydotool', 'type', '--', to_type])
            elif token in STRING_AUTOTYPE_TOKENS:
                to_type = STRING_AUTOTYPE_TOKENS[token]
                call(['ydotool', 'type', '--', to_type])
            elif token in AUTOTYPE_TOKENS:
                cmd = ['ydotool'] + AUTOTYPE_TOKENS[token]
                call(cmd)
            else:
                dmenu_err(f"Unsupported auto-type token (ydotool): \"{token}\"")
                return
        else:
            call(['ydotool', 'type', '--', token])


def type_entry_wtype(entry, tokens):
    """Auto-type entry entry using wtype

    """
    from .tokens_wtype import AUTOTYPE_TOKENS
    for token, special in tokens:
        if special:
            cmd = token_command(token)
            if callable(cmd):
                to_type = cmd(entry)  # pylint: disable=not-callable
                if to_type is not None:
                    call(['wtype', '--', to_type])
            elif token in PLACEHOLDER_AUTOTYPE_TOKENS:
                to_type = PLACEHOLDER_AUTOTYPE_TOKENS[token](entry)
                if to_type:
                    call(['wtype', '--', to_type])
            elif token in STRING_AUTOTYPE_TOKENS:
                to_type = STRING_AUTOTYPE_TOKENS[token]
                call(['wtype', '--', to_type])
            elif token in AUTOTYPE_TOKENS:
                cmd = ['wtype', '-k', AUTOTYPE_TOKENS[token]]
                call(cmd)
            else:
                dmenu_err(f"Unsupported auto-type token (wtype): \"{token}\"")
                return
        else:
            call(['wtype', '--', token])

def type_text(data):
    """Type the given text data

    """
    library = 'pynput'
    if keepmenu.CONF.has_option('database', 'type_library'):
        library = keepmenu.CONF.get('database', 'type_library')
    if library == 'xdotool':
        call(['xdotool', 'type', '--', data])
    elif library == 'ydotool':
        call(['ydotool', 'type', '--', data])
    elif library == 'wtype':
        call(['wtype', '--', data])
    else:
        try:
            from pynput import keyboard
        except ModuleNotFoundError:
            return
        kbd = keyboard.Controller()
        try:
            kbd.type(data)
        except kbd.InvalidCharacterException:
            dmenu_err("Unable to type string...bad character.\n"
                      "Try setting `type_library = xdotool` in config.ini")



"""Read and copy Keepass database entries using dmenu or rofi

"""
from copy import deepcopy
from datetime import datetime, timedelta
import errno
import functools
from multiprocessing import Process
import os
from os.path import expanduser, realpath
import shlex
import subprocess
import sys
from threading import Timer

import construct
import keepmenu
from keepmenu.edit import add_entry, edit_entry, manage_groups
from keepmenu.menu import dmenu_err, dmenu_select
from keepmenu.type import type_entry, type_text
from keepmenu.view import view_all_entries, view_entry


class DataBase():  # pylint: disable=too-few-public-methods
    """Define a database class for clearer reference to variables

        dbase - string, filename
        kfile - string, filename
        pword - string, password
        atype - string, autotype sequence
        kpo - PyKeePass object
        is_active - bool, is this the currently active database

    """
    def __init__(self, dbase=None, kfile=None, pword=None, atype=None, kpo=None):
        # pylint: disable=too-many-arguments
        self.dbase = expanduser('' if dbase is None else dbase)
        self.dbase = realpath(self.dbase) if self.dbase else ''
        self.kfile = expanduser('' if kfile is None else kfile)
        self.kfile = realpath(self.kfile) if self.kfile else ''
        self.pword = pword
        self.atype = atype
        self.kpo = kpo
        self.is_active = False

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


def get_databases():
    """Read databases from config

    Returns: [DataBase obj, DataBase obj2,...]
             If not specified in the config, the value will be None
             If database name is None, an error has occurred

    """
    dargs = keepmenu.CONF.items('database')
    args_dict = dict(dargs)
    dbases = [i for i in args_dict if i.startswith('database')]
    dbs = []
    for dbase in dbases:
        dbn = args_dict[dbase]
        idx = dbase.rsplit('_', 1)[-1]
        try:
            keyfile = args_dict[f'keyfile_{idx}']
        except KeyError:
            keyfile = None
        try:
            passw = args_dict[f'password_{idx}']
        except KeyError:
            passw = None
        try:
            autotype = args_dict[f'autotype_default_{idx}']
        except KeyError:
            autotype = None
        try:
            cmd = args_dict[f'password_cmd_{idx}']
            res = subprocess.run(shlex.split(cmd),
                                 capture_output=True,
                                 check=False,
                                 encoding=keepmenu.ENC)
            if res.stderr:
                dmenu_err(f"Password command error: {res.stderr}")
                sys.exit()
            else:
                passw = res.stdout.rstrip('\n') if res.stdout else passw
        except KeyError:
            pass
        if dbn:
            dbo = DataBase(dbase=dbn, kfile=keyfile, pword=passw, atype=autotype)
            dbs.append(dbo)

    return dbs


def get_database(open_databases=None, **kwargs):
    # pylint: disable=too-many-statements,too-many-branches
    """Read databases/keyfile/autotype from config, CLI, or ask for user input.

    Args: open_databases - list [DataBase1, DataBase2,...]
          kwargs - possibly 'database', 'keyfile'
    Returns: DataBase obj or None on error selecting database or password
             open_databases - list [DataBase1, DataBase2,...]
                (open_databases is managed in get_database but stored in
                DmenuRunner for persistence instead of using a global var)

    """
    dbs_cfg = get_databases()
    dbs_cfg_n = [i.dbase for i in dbs_cfg]
    open_databases = open_databases or {}
    clidb = DataBase(dbase=kwargs.get('database'),
                     kfile=kwargs.get('keyfile'),
                     atype=kwargs.get('autotype'))
    if not dbs_cfg and not clidb.dbase and not open_databases:
        # First run  database opening
        res = get_initial_db()
        if res is True:
            db_, open_databases = get_database()
            dbs = [db_]
        else:
            return None, open_databases
    elif clidb.dbase:
        # Prefer db and autotype passed via cli
        db_ = [i for i in open_databases.values() if i.dbase == clidb.dbase]
        if db_:
            dbs = [deepcopy(db_[0])]
            if clidb.atype:
                dbs[0].atype = clidb.atype
        else:
            dbs = [deepcopy(clidb)]
            # Use existing keyfile if available
            if not dbs[0].kfile and dbs[0].dbase in dbs_cfg_n:
                dbs[0].kfile = dbs_cfg[dbs_cfg_n.index(dbs[0].dbase)].kfile
            # Use existing password if available
            if dbs[0].pword is None and dbs[0].dbase in dbs_cfg_n:
                dbs[0].pword = dbs_cfg[dbs_cfg_n.index(dbs[0].dbase)].pword
    elif clidb.atype and open_databases:
        # If only autotype is passed, use current db
        db_ = [i for i in open_databases.values() if i.is_active is True][0]
        dbs = [deepcopy(db_)]
        dbs[0].atype = clidb.atype
    elif open_databases:
        # if there are dbs already open, make a list of those + dbs from config.ini
        dbs = [deepcopy(i) for i in open_databases.values()]
        for db_ in dbs_cfg:
            if db_.dbase not in open_databases:
                dbs.append(deepcopy(db_))
    else:
        dbs = dbs_cfg
    if len(dbs) > 1:
        inp = "\n".join(i.dbase for i in dbs)
        sel = dmenu_select(len(dbs), "Select Database", inp=inp)
        dbs = [i for i in dbs if i.dbase == sel]
        if not sel or not dbs:
            return None, open_databases
    if dbs[0].pword is None:
        dbs[0].pword = get_passphrase()
        if dbs[0].pword is None:
            return None, open_databases
    for db_ in open_databases.values():
        db_.is_active = False
    if dbs[0].dbase not in open_databases:
        open_databases[dbs[0].dbase] = deepcopy(dbs[0])
    if dbs[0].dbase in dbs_cfg_n:
        db_cfg_atype = dbs_cfg[dbs_cfg_n.index(dbs[0].dbase)].atype
        open_databases[dbs[0].dbase].atype = db_cfg_atype
        if not dbs[0].atype:
            dbs[0].atype = db_cfg_atype
    if clidb.atype:
        dbs[0].atype = clidb.atype
    open_databases[dbs[0].dbase].is_active = True
    return dbs[0], open_databases


def get_initial_db():
    """Ask for and set initial database name and keyfile if not entered in config file

    """
    db_name = dmenu_select(0, "Enter path to existing "
                              "Keepass database. ~/ for $HOME is ok")
    if not db_name:
        dmenu_err("No database entered. Try again.")
        return False
    keyfile_name = dmenu_select(0, "Enter path to keyfile. ~/ for $HOME is ok")
    with open(keepmenu.CONF_FILE, 'w', encoding=keepmenu.ENC) as conf_file:
        keepmenu.CONF.set('database', 'database_1', db_name)
        if keyfile_name:
            keepmenu.CONF.set('database', 'keyfile_1', keyfile_name)
        keepmenu.CONF.write(conf_file)
    return True


def get_entries(dbo):
    """Open keepass database and return the PyKeePass object

        Args: dbo: DataBase object
        Returns: PyKeePass object or None

    """
    from pykeepass import PyKeePass  # pylint: disable=import-outside-toplevel
    if dbo.dbase is None:
        return None
    try:
        kpo = PyKeePass(dbo.dbase, dbo.pword, keyfile=dbo.kfile)
    except (FileNotFoundError, construct.core.ChecksumError) as err:
        if str(err.args[0]).startswith("wrong checksum"):
            dmenu_err("Invalid Password or keyfile")
            return None
        try:
            if err.errno == errno.ENOENT:
                if not os.path.isfile(dbo.dbase):
                    dmenu_err("Database does not exist. Check path and filename.")
                elif not os.path.isfile(dbo.kfile):
                    dmenu_err("Keyfile does not exist. Check path and filename.")
        except AttributeError:
            pass
        return None
    except Exception as err:  # pylint: disable=broad-except
        dmenu_err(f"Error: {err}")
        return None
    return kpo


def get_passphrase():
    """Get a database password from dmenu or pinentry

    Returns: string

    """
    pinentry = keepmenu.CONF.get("dmenu", "pinentry", fallback=None)
    if pinentry:
        password = ""
        res = subprocess.run(pinentry,
                             capture_output=True,
                             check=False,
                             encoding=keepmenu.ENC,
                             input='setdesc Enter database password\ngetpin\n')
        if res.stdout:
            pin = res.stdout.split("\n")[2]
            if pin.startswith("D "):
                password = pin.split("D ")[1]
    else:
        password = dmenu_select(0, "Password")
    return None or password


def get_expiring_entries(entries):
    """Return a list of expired entries (if they can expire)

    """
    expiring = []
    for entry in entries:
        if entry.expires is True and \
                entry.expiry_time.timestamp() < (datetime.now() + timedelta(3)).timestamp():
            expiring.append(entry)
    return expiring


class DmenuRunner(Process):
    """Listen for dmenu calling event and run keepmenu

    Args: server - Server object
          kpo - Keepass object
    """

    def __init__(self, server, **kwargs):
        Process.__init__(self)
        self.server = server
        self.database, self.open_databases = get_database(**kwargs)
        if self.database:
            self.database.kpo = get_entries(self.database)
        if not self.database or not self.database.kpo:
            self.server.kill_flag.set()
            sys.exit()
        self.expiring = get_expiring_entries(self.database.kpo.entries)
        self.prev_entry = None

    def _set_timer(self):
        """Set inactivity timer

        """
        # pylint: disable=attribute-defined-outside-init
        self.cache_timer = Timer(keepmenu.CACHE_PERIOD_MIN * 60, self.cache_time)
        self.cache_timer.daemon = True
        self.cache_timer.start()

    def run(self):
        while True:
            self.server.start_flag.wait()
            if self.server.kill_flag.is_set():
                break
            if not self.database or not self.database.kpo:
                pass
            elif self.server.args_flag.is_set():
                dargs = self.server.get_args()
                self.menu_open_another_database(**dargs)
                self.server.args_flag.clear()
            else:
                self.dmenu_run()
            if self.server.cache_time_expired.is_set():
                self.server.kill_flag.set()
            if self.server.kill_flag.is_set():
                break
            self.server.start_flag.clear()

    def cache_time(self):
        """Kill keepmenu daemon when cache timer expires

        """
        self.server.cache_time_expired.set()
        self.server.args_flag.clear()
        if not self.server.start_flag.is_set():
            self.server.kill_flag.set()
            self.server.start_flag.set()

    def dmenu_run(self):
        """Run dmenu with the given list of Keepass Entry objects

        If 'hide_groups' is defined in config.ini, hide those from main and
        view/type all views.

        Args: self.database.kpo - Keepass object

        Note: I had to reload the kpo object after every save to prevent being
        affected by the gibberish password bug in pykeepass:
        https://github.com/pschmitt/pykeepass/issues/43

        Once this is fixed, the extra calls to self.database.kpo = get_entries... can be
        deleted

        """
        try:
            self.cache_timer.cancel()
        except AttributeError:
            pass
        self._set_timer()
        if keepmenu.CONF.has_option("database", "hide_groups"):
            hid_groups = keepmenu.CONF.get("database", "hide_groups").split("\n")
            # Validate ignored group names in config.ini
            hid_groups = [i for i in hid_groups if i in
                          [j.name for j in self.database.kpo.groups]]
        else:
            hid_groups = []

        filtered_entries = [
            i for i in self.database.kpo.entries if not
            any(j in "/".join(i.path[:-1]) for j in hid_groups)
        ]
        options = {
            'View/Type Individual entries':
                functools.partial(self.menu_view_type_individual_entries, hid_groups),
            'View previous entry': self.menu_view_previous_entry,
            f'Edit expiring/expired passwords ({len(self.expiring)})':
                functools.partial(self.menu_edit_entries, self.expiring),
            'Edit entries': functools.partial(self.menu_edit_entries, self.database.kpo.entries),
            'Add entry': self.menu_add_entry,
            'Manage groups': self.menu_manage_groups,
            'Reload database': self.menu_reload_database,
            'Open another database': self.menu_open_another_database,
            'Kill Keepmenu daemon': self.menu_kill_daemon,
        }
        if self.prev_entry is None:
            del options['View previous entry']
        if len(self.expiring) == 0:
            del options['Edit expiring/expired passwords (0)']

        sel = view_all_entries(list(options), filtered_entries, self.database.dbase)
        if not sel:
            return
        if sel in options:
            func = options[sel]
            func()
        else:
            try:
                entry = filtered_entries[int(sel.split('-', 1)[0])]
            except (ValueError, TypeError):
                return
            type_entry(entry, self.database.atype)
            self.prev_entry = entry
        # Reset database autotype in between runs
        cur_db = [i for i in self.open_databases.values() if i.is_active is True][0]
        self.database.atype = cur_db.atype

    def menu_view_type_individual_entries(self, hid_groups):
        """Process menu entry - View/Type individual entries

        """
        options = []
        filtered_entries = [
            i for i in self.database.kpo.entries if not
            any(j in "/".join(i.path[:-1]) for j in hid_groups)
        ]
        sel = view_all_entries(options, filtered_entries, self.database.dbase)
        try:
            entry = filtered_entries[int(sel.split('-', 1)[0])]
        except (ValueError, TypeError):
            return
        text = view_entry(entry)
        type_text(text)
        self.prev_entry = entry

    def menu_view_previous_entry(self):
        """Process menu entry - View previous entry

        """
        assert self.prev_entry is not None
        text = view_entry(self.prev_entry)
        type_text(text)

    def menu_edit_entries(self, entries):
        """Process menu entry - Edit individual entries

        """
        options = []
        sel = view_all_entries(options, entries, self.database.dbase)
        try:
            entry = entries[int(sel.split('-', 1)[0])]
        except (ValueError, TypeError):
            return
        edit = True
        while edit is True:
            edit = edit_entry(self.database.kpo, entry)
        self.database.kpo.save()
        self.database.kpo = get_entries(self.database)
        self.prev_entry = entry if edit != "del" else None

    def menu_add_entry(self):
        """Process menu entry - Add entry

        """
        entry = add_entry(self.database.kpo)
        if entry:
            self.database.kpo.save()
            self.database.kpo = get_entries(self.database)
            self.prev_entry = entry

    def menu_manage_groups(self):
        """Process menu entry - manage groups

        """
        group = manage_groups(self.database.kpo)
        if group:
            self.database.kpo.save()
            self.database.kpo = get_entries(self.database)

    def menu_reload_database(self):
        """Process menu entry - Reload database

        """
        self.database.kpo = get_entries(self.database)
        if not self.database.kpo:
            return
        self.expiring = get_expiring_entries(self.database.kpo.entries)
        self.dmenu_run()

    def menu_open_another_database(self, **kwargs):
        """Process menu entry - Open different database

        Args: kwargs - possibly 'database', 'keyfile', 'autotype'

        """
        prev_db, prev_open = self.database, deepcopy(self.open_databases)
        self.database, self.open_databases = get_database(self.open_databases, **kwargs)
        if self.database is None:
            self.database, self.open_databases = prev_db, prev_open
            return
        if not self.database.kpo:
            self.database.kpo = get_entries(self.database)
            if self.database.kpo is None:
                self.database, self.open_databases = prev_db, prev_open
                return
        self.expiring = get_expiring_entries(self.database.kpo.entries)
        self.dmenu_run()

    def menu_kill_daemon(self):
        """Process menu entry - Kill keepmenu daemon

        """
        try:
            self.server.kill_flag.set()
        except (EOFError, IOError):
            return


# vim: set et ts=4 sw=4 :



""" TOTP generation

"""
import base64
import hmac
import struct
import time
from urllib import parse


def hotp(key, counter, digits=6, digest='sha1', steam=False):
    """ Generates HMAC OTP.  Taken from https://github.com/susam/mintotp

    Args: key - Secret key
          counter - Moving factor
          digits - The number of characters/digits that the otp should have
          digest - Algorithm to use to generate the otp
          steam - whether or not to use steam settings

    Returns: otp

    """
    key = base64.b32decode(key.upper() + '=' * ((8 - len(key)) % 8))
    counter = struct.pack('>Q', counter)
    mac = hmac.new(key, counter, digest).digest()
    offset = mac[-1] & 0x0f
    binary = struct.unpack('>L', mac[offset:offset + 4])[0] & 0x7fffffff
    code = ''

    if steam:
        chars = '23456789BCDFGHJKMNPQRTVWXY'
        full_code = int(binary)
        for _ in range(digits):
            code += chars[full_code % len(chars)]
            full_code //= len(chars)
    else:
        code = str(binary)[-digits:].rjust(digits, '0')

    return code


def totp(key, time_step=30, digits=6, digest='sha1', steam=False):
    """ Generates Time Based OTP

    Args: key - Secret key
          counter - The length of time in seconds each otp is valid for
          digits - The number of characters/digits that the otp should have
          digest - Algorithm to use to generate the otp
          steam - whether or not to use steam settings

    Returns: otp

    """
    return hotp(key, int(time.time() / time_step), digits, digest, steam)


def gen_otp(otp_url):
    """ Generates one time password

    Args: otp_url - KeePassXC url encoding with information on how to generate otp
    Returns: otp

    """
    parsed_otp_url = parse.urlparse(otp_url)
    query_string = parse.parse_qs(parsed_otp_url.query)

    for required in ('secret', 'period', 'digits'):
        if required not in query_string:
            return ''

    try:
        steam = query_string['encoder'][0] == "steam"
    except KeyError:
        steam = False

    return totp(
        query_string['secret'][0], int(query_string['period'][0]),
        int(query_string['digits'][0]), 'sha1' if 'algorihm'
        not in query_string else query_string['algorithm'][0].lower(), steam)


def get_otp_url(kp_entry):
    """ Shim to return otp url from KeePass entry
    This is required to fully support pykeepass>=4.0.0
    "otp" was upgraded to a reserved property in pykeepass==4.0.3

    Args: kp_entry - KeePassXC entry
    Returns: otp url string or None

    """
    if hasattr(kp_entry, "otp"):
        return kp_entry.deref("otp")
    else:
        return kp_entry.get_custom_property("otp")



"""Main entrypoint and CLI parsing

"""
import argparse
import configparser
from contextlib import closing
from multiprocessing import Event, Process, Pipe
from multiprocessing.managers import BaseManager
import os
from os.path import exists, expanduser
import random
import socket
import string
from subprocess import call

import keepmenu
from keepmenu.keepmenu import DmenuRunner


def find_free_port():
    """Find random free port to use for BaseManager server

    Returns: int Port

    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(('127.0.0.1', 0))  # pylint:disable=no-member
        return sock.getsockname()[1]  # pylint:disable=no-member


def get_auth():
    """Generate and save port and authkey to ~/.cache/.keepmenu-auth

    Returns: int port, bytestring authkey

    """
    auth = configparser.ConfigParser()
    if not exists(keepmenu.AUTH_FILE):
        fd_ = os.open(keepmenu.AUTH_FILE, os.O_WRONLY | os.O_CREAT, 0o600)
        with open(fd_, 'w', encoding=keepmenu.ENC) as a_file:
            auth.set('DEFAULT', 'port', str(find_free_port()))
            auth.set('DEFAULT', 'authkey', random_str())
            auth.write(a_file)
    try:
        auth.read(keepmenu.AUTH_FILE)
        port = auth.get('DEFAULT', 'port')
        authkey = auth.get('DEFAULT', 'authkey').encode()
    except (configparser.NoOptionError, configparser.MissingSectionHeaderError):
        os.remove(keepmenu.AUTH_FILE)
        print("Cache file was corrupted. Stopping all instances. Please try again")
        call(["pkill", "keepmenu"])  # Kill all prior instances as well
        return None, None
    return int(port), authkey


def random_str():
    """Generate random auth string for BaseManager

    Returns: string

    """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(15))


def client():
    """Define client connection to server BaseManager

    Returns: BaseManager object
    """
    port, auth = get_auth()
    mgr = BaseManager(address=('', port), authkey=auth)
    mgr.register('set_event')
    mgr.register('get_pipe')
    mgr.register('read_args_from_pipe')
    mgr.connect()
    return mgr


class Server(Process):  # pylint: disable=too-many-instance-attributes
    """Run BaseManager server to listen for dmenu calling events

    """
    def __init__(self):
        Process.__init__(self)
        self.port, self.authkey = get_auth()
        self.start_flag = Event()
        self.kill_flag = Event()
        self.cache_time_expired = Event()
        self.args_flag = Event()
        self.start_flag.set()
        self.args = None
        self._parent_conn, self._child_conn = Pipe(duplex=False)

    def run(self):
        _ = self.server()
        self.kill_flag.wait()

    def _get_pipe(self):
        return self._child_conn

    def get_args(self):
        """ Reads aruments sent by the client to the server

        """
        return self._parent_conn.recv()

    def server(self):
        """Set up BaseManager server

        """
        mgr = BaseManager(address=('127.0.0.1', self.port),
                          authkey=self.authkey)
        mgr.register('set_event', callable=self.start_flag.set)
        mgr.register('get_pipe', callable=self._get_pipe)
        mgr.register('read_args_from_pipe', callable=self.args_flag.set)
        mgr.start()  # pylint: disable=consider-using-with
        return mgr


def run(**kwargs):
    """Start the background Manager and Dmenu runner processes.

    """
    server = Server()
    dmenu = DmenuRunner(server, **kwargs)
    dmenu.daemon = True
    server.start()
    dmenu.start()
    server.join()
    if exists(expanduser(keepmenu.AUTH_FILE)):
        os.remove(expanduser(keepmenu.AUTH_FILE))


def main():
    """Main script entrypoint

    """
    parser = argparse.ArgumentParser(
        description="Dmenu/Rofi frontend for Keepass databases")

    parser.add_argument(
        "-a",
        "--autotype",
        type=str,
        required=False,
        help="Override autotype sequence in config.ini",
    )

    parser.add_argument(
        "-d",
        "--database",
        type=str,
        required=False,
        help="File path to a database to open, skipping the database selection menu",
    )

    parser.add_argument(
        "-k",
        "--keyfile",
        type=str,
        required=False,
        help="File path of the keyfile needed to open the database specified by --database/-d",
    )

    args = vars(parser.parse_args())

    args = args if any(args.values()) else {}

    try:
        manager = client()
        conn = manager.get_pipe()  # pylint: disable=no-member
        if args:
            conn.send(args)
            manager.read_args_from_pipe()  # pylint: disable=no-member
        manager.set_event()  # pylint: disable=no-member
    except ConnectionRefusedError:
        run(**args)


if __name__ == '__main__':
    main()

# vim: set et ts=4 sw=4 :



#!/usr/bin/env python3  # pylint: disable=missing-docstring

from setuptools import setup

setup(name="keepmenu",
      version="1.3.1",
      description="Dmenu frontend for Keepass databases",
      long_description=open('README.md', 'rb').read().decode('utf-8'),
      long_description_content_type="text/markdown",
      author="Scott Hansen",
      author_email="firecat4153@gmail.com",
      url="https://github.com/firecat53/keepmenu",
      download_url="https://github.com/firecat53/keepmenu/tarball/1.3.1",
      packages=['keepmenu'],
      entry_points={
          'console_scripts': ['keepmenu=keepmenu.__main__:main']
      },
      data_files=[('share/doc/keepmenu', ['README.md', 'LICENSE',
                                          'config.ini.example']),
                  ('share/doc/keepmenu/docs', ['docs/install.md',
                                               'docs/configure.md',
                                               'docs/usage.md']),
                  ('share/man/man1', ['keepmenu.1'])],
      install_requires=["pynput", "pykeepass>=4.0.0"],
      license="GPL3",
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: End Users/Desktop',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Utilities',
      ],
      keywords=("dmenu keepass keepassxc"),
      )
